import zlib
import base64
import io

import pandas as pd
from pydantic import (
    BaseModel,
    field_serializer,
    field_validator,
    ConfigDict,
)

SUPPORTED_FILETYPES = [
    "csv",
]


class DataSource(BaseModel):
    file_path: str
    raw: str | pd.DataFrame

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_serializer("raw")
    def serialize_raw(self, raw: pd.DataFrame):
        return self._dump_data_frame(raw)

    def __eq__(self, other):
        if not isinstance(other, DataSource):
            return False
        return self.file_path == other.file_path

    @field_validator("raw")
    @classmethod
    def validate_raw(cls, value: str | pd.DataFrame) -> pd.DataFrame:
        if isinstance(value, pd.DataFrame):
            return value
        try:
            return cls._load_data_frame(value)
        except Exception:
            raise ValueError("unable to load a `pandas.DataFrame` object from raw")

    @staticmethod
    def _dump_data_frame(data_frame: pd.DataFrame) -> str:
        data_frame_bytes = data_frame.to_parquet()
        compressed = zlib.compress(data_frame_bytes)
        compressed_b64 = base64.b64encode(compressed)
        compressed_b64_string = compressed_b64.decode()
        return compressed_b64_string

    @staticmethod
    def _load_data_frame(compressed: str) -> pd.DataFrame:
        compressed_b64_bytes = compressed.encode()
        compressed_b64 = base64.b64decode(compressed_b64_bytes)
        uncompressed = zlib.decompress(compressed_b64)
        data_frame = pd.read_parquet(io.BytesIO(uncompressed))
        return data_frame



import pytest

from entropic.process import Pipeline
from entropic.process import exceptions


def test_required_definitions():
    with pytest.raises(exceptions.PipelineSetupError) as error:
        Pipeline()
    assert str(error.value) == "can't instantiate Pipeline directly"

    with pytest.raises(exceptions.PipelineSetupError) as error:

        class TestNoExtract(Pipeline):
            source_paths = ["test/path"]

    assert str(error.value) == "either 'extract_with' or 'extract' must be defined"

    with pytest.raises(exceptions.PipelineSetupError) as error:

        class TestNoSource(Pipeline):
            extract_with = lambda x: x  # noqa: E731

    assert (
        str(error.value)
        == "either 'source_paths' or 'get_source_paths' must be defined"
    )

    with pytest.warns() as warnings:

        class TestSourceAndFilePaths(Pipeline):
            source_paths = ["test/path"]
            extract_with = lambda x: x  # noqa: E731

            def get_source_paths(self):
                return []

        class Process(Pipeline):
            source_paths = ["test/path"]
            extract_with = lambda x: x  # noqa: E731

            def extract(self):
                return 1

    assert len(warnings) == 2
    assert (
        str(warnings[0].message)
        == "both 'source_paths' and 'get_source_paths' defined, ignoring 'source_paths'"
    )
    assert (
        str(warnings[1].message)
        == "both 'extract_with' and 'extract' are defined, ignoring 'extract_with'"
    )

    with pytest.raises(TypeError) as error:

        class BadSourcePaths(Pipeline):
            source_paths = "not-a-list"
            extract_with = lambda x: x  # noqa: E731

        BadSourcePaths()

    assert str(error.value) == "'source_paths' must be a list of path-like objects"

    with pytest.raises(TypeError) as error:

        class BadSourcePaths(Pipeline):
            source_paths = [1, 2, 3]
            extract_with = lambda x: x  # noqa: E731

        BadSourcePaths()

    assert str(error.value) == "'source_paths' must be a list of path-like objects"


def test_default_functions():
    def my_extract_function(filename):
        return filename

    class TestDefaultExtract(Pipeline):
        source_paths = ["test/path"]
        extract_with = my_extract_function

    assert TestDefaultExtract.extract_with == my_extract_function

    class TestCustomFilepaths(Pipeline):
        extract_with = my_extract_function

        def get_source_paths(self):
            return ["file"]

    assert TestCustomFilepaths.source_paths == []



import pytest
import pandas as pd
from pydantic import ValidationError

from entropic.sources.fields import DataSource

TEST_DATA_FRAME = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})


def test_raw_validation_and_serialization():
    with pytest.raises(ValidationError) as error:
        DataSource(file_path="", raw="invalid df")
    assert "unable to load a `pandas.DataFrame` object from raw" in str(error.value)

    valid = DataSource(file_path="some_path", raw=TEST_DATA_FRAME)
    assert isinstance(valid.model_dump()["raw"], str)


def test_comparisons():
    assert DataSource(file_path="some_path", raw=TEST_DATA_FRAME) == DataSource(
        file_path="some_path", raw=TEST_DATA_FRAME
    )
    assert DataSource(file_path="some_path", raw=TEST_DATA_FRAME) != DataSource(
        file_path="other_path", raw=TEST_DATA_FRAME
    )
    assert (
        DataSource(file_path="some_path", raw=TEST_DATA_FRAME)
        != "not a data source instance"
    )


def test_dump_and_load_compressed():
    compressed = DataSource._dump_data_frame(TEST_DATA_FRAME)
    assert (
        compressed
        == "eJy1VsuO4kYUNWARImXRGsUIpF4gK0FZ0B3shplhNFmU6QZMgBnoaWwcRS2/2hhs4/jRhm7xB1lkkezzOVlmkUU+IB+SqjLvZiKNlC4h6ta5t849t1zX8nswYCiSKlBfdeBEvCCIk7cJIh7J1UwR1AsqW4L+EypNpU9zmbUnl1nHZolvjqBwfF5AMJlIpJLFvxOnFJmvEumTfI5UZxZDJbNk9o9E9s9EsVXMfApxHskhqCQSd0KhgMMqUqtI8qNVkBva1E6yp+hBFf+Q+1WwO1X8lSr+mvoU6mN15N9WiVzaV8e6LVMkBL5Oxse1NVkiS+ZP86XnPdLnLjX7SypLFjPZ31NfIncu7cqOJvu538hH2nQ0fX4Lc4a249NvCj880lOIQYv2ZMfQ6VKBdmRbh4ATWhZc+YHsBXBZxvbMhSaLTR2ZzPJHuIj5bjG5vqLdZ7kzdUu73cdiWbfBwkUgHTqmOtNiBaHtLjaOmTLR1QDhth7ImhzIEH2kdQeGm46BQm4+NM5e08sdMfsqEMjQhzo26IES0wleVp7q2MA7MlApy1JhLxF7NBH7PyTC5Xm6HMw8fASWqXiyt0B73IXsebMI7brXPd+cOQhlKuflc4ZeblPvONlz5vyCXhK5L8Bg8E54E7dG7ufMt2hE4BIA0Afx+B6ACuAM0ADAAHUDYxy354/QfLXxX0X1eL1mwX4ezRyejdV6HTRClNFIqE5HYheaN3A/94q3rVBqDisi27tXWsNAcbrmO5NT9EXbk2EsP5kZvNMeK7Zm8WZk8HZ1rAg3KCbUhLnfqQNTvRiMVacPMRCt1vdqnX/JX/I+X28/aM1hhHi6IqMh/4it+ZrAhCIb55bETc5wJDAWilWcoa/UUT7JlYT5VGSP5G024P7Gg3gxqKpN5AOmJlRdyD+VrnFuGMtE+rBW1kUO8/J2zZTt4USrY38giYMx5CiPrqHelm9KQnWisANXsVXMNxwOmp0Phnl3oF1dwPhG8Gp7HrGvG+e1ZWHoS43aph7ehud7eWV+TDc863Jv0sd+WHugtqw7rWVFEtJVb7uK02d7xzQ3q4zSnNc6de6pFvO/tPDPpkW8Rs+tt5CERlm54NEzg9rm7gjeI9WZ4hyQE9rte22BeVhJbD/IQi3E+iZX5c4EhN3rdm1fo2SpTs9V2Arm6JqVeWfSNe/666vO4fvOob/eJWyy1+j21/H954xtK2yahoPPOlrgZXcdD7j+iqSBSUewa0B3tXndjxwY49nY5IsH5sdnH0fifoSN1lqFbvnj8R2RK7iy91OoB2eq657hl0xh9RIpxK+XfOmUIOCP0D4jiPfwy+df7J032w=="
    )

    decompressed = DataSource._load_data_frame(compressed)
    assert isinstance(decompressed, pd.DataFrame)
    assert (
        from_load.all() == from_test.all()
        for from_load, from_test in zip(decompressed.all(), TEST_DATA_FRAME.all())
    )



import pandas as pd

from entropic.process import Pipeline
from entropic import results


class Process(Pipeline):
    source_paths = ["../tests/mocks/"]
    extract_with = pd.read_csv


p = Process()
p.run()


if __name__ == "__main__":
    for iteration in results.all:
        for sample in iteration.samples:
            print(sample.data.raw.head())



import os
import warnings
from typing import final, Callable
from pathlib import Path

from entropic.sources import Iteration, Sample
from entropic.sources.fields import DataSource

from entropic.process.exceptions import PipelineSetupError


class PipelineMeta(type):
    def __new__(cls, name, bases, attrs):
        if not bases:
            # Pipeline instantiation error handled in Pipeline.__init__
            return super().__new__(cls, name, bases, attrs)

        if not (attrs.get("source_paths") or attrs.get("get_source_paths")):
            raise PipelineSetupError(
                "either 'source_paths' or 'get_source_paths' must be defined"
            )
        if not (attrs.get("extract_with") or attrs.get("extract")):
            raise PipelineSetupError(
                "either 'extract_with' or 'extract' must be defined"
            )

        if attrs.get("source_paths") and attrs.get("get_source_paths"):
            warnings.warn(
                "both 'source_paths' and 'get_source_paths' defined, ignoring 'source_paths'",
                stacklevel=2,
            )
        if attrs.get("extract_with") and attrs.get("extract"):
            warnings.warn(
                "both 'extract_with' and 'extract' are defined, ignoring 'extract_with'",
                stacklevel=2,
            )

        if extract_with := attrs.get("extract_with"):
            attrs["extract_with"] = staticmethod(extract_with)

        return super().__new__(cls, name, bases, attrs)


class Pipeline(metaclass=PipelineMeta):
    iteration = Iteration
    source_paths: list[Path | str] = []
    extract_with: Callable

    def __init__(self):
        if type(self) == Pipeline:
            raise PipelineSetupError("can't instantiate Pipeline directly")

        if self.source_paths:
            error_message = "'source_paths' must be a list of path-like objects"
            if not isinstance(self.source_paths, list):
                raise TypeError(error_message)
            try:
                self.source_paths = [Path(path) for path in self.source_paths]
            except TypeError as ex:
                raise TypeError(error_message) from ex

    def get_source_paths(self):
        return self.source_paths

    def get_iteration(self):
        return self.iteration

    def get_files_from_path(self, path):
        return [Path(path, file) for file in os.listdir(path)]

    def extract(self, file_path) -> Sample:
        data_source_data = self.extract_with(file_path)
        return Sample(data=DataSource(file_path=str(file_path), raw=data_source_data))

    @final
    def extract_all_source_paths(self):
        for source_path in self.get_source_paths():
            instance = self.get_iteration().get_or_create(source_path=str(source_path))
            for file_path in self.get_files_from_path(source_path):
                sample = self.extract(file_path)
                instance.upsert_sample(sample=sample)
            instance.save()

    @final
    def run(self):
        self.extract_all_source_paths()
        # TODO: add load methods
        return



from typing import Optional

from tinydb import TinyDB, where, Query

from entropic.db.base import BaseHandler


class Handler(BaseHandler):
    PATH = "./.entropic-db"

    def __init__(self, path=None):
        if path:
            self.PATH = path

    @property
    def database(self) -> TinyDB:
        return TinyDB(self.PATH)

    @staticmethod
    def _kwargs_to_query(kwargs: dict) -> Optional[Query]:
        if not (items := list(kwargs.items())):
            return None
        query = where(items[0][0]) == items[0][1]
        for field, value in items[1:]:
            query &= Query()[field] == value
        return query

    def get(self, **kwargs):
        id = kwargs.pop("id", None)
        query = self._kwargs_to_query(kwargs)
        return self.database.get(query, doc_id=id)

    def all(self):
        return self.database.all()

    def filter(self, **kwargs):
        if not kwargs:
            return self.all()
        query = self._kwargs_to_query(kwargs)
        return self.database.search(query)

    def insert_one(self, document):
        return self.database.insert(document)

    def get_or_create(self, **kwargs):
        if not (item := self.get(**kwargs)):
            item = kwargs
            self.insert_one(item)

        return item

    def upsert(self, document, key: Optional[dict] = None):
        if not key:
            key = {"key": "id", "value": document.get("id")}
        return self.database.upsert(document, where(key["key"]) == key["value"])



import abc
from typing import Collection


class BaseHandler(abc.ABC):
    @property
    @abc.abstractmethod
    def database(self):
        ...

    @abc.abstractmethod
    def get(self, **kwargs):
        ...

    @abc.abstractmethod
    def all(self):
        ...

    @abc.abstractmethod
    def filter(self, **kwargs):
        ...

    @abc.abstractmethod
    def insert_one(self, document: Collection):
        ...

    @abc.abstractmethod
    def get_or_create(self, **kwargs):
        ...



import os

import pytest
from tinydb import TinyDB
from tinydb.queries import QueryInstance

from entropic.db import TinyDBHandler


@pytest.fixture
def database():
    DB_PATH = ".entropic-db-2"
    yield TinyDBHandler(path=DB_PATH)
    try:
        os.remove(DB_PATH)
    except FileNotFoundError:
        pass


def test_initialization(database):
    assert isinstance(database.database, TinyDB)


def test_kwargs_to_db_filter():
    assert TinyDBHandler._kwargs_to_query({}) is None

    kwargs = {"field_1": "value_1"}
    query = TinyDBHandler._kwargs_to_query(kwargs)
    assert isinstance(query, QueryInstance)
    assert query == QueryInstance(test=None, hashval=("==", ("field_1",), "value_1"))

    kwargs["field_2"] = "value_2"
    query = TinyDBHandler._kwargs_to_query(kwargs)
    assert isinstance(query, QueryInstance)
    assert query == QueryInstance(
        test=None,
        hashval=(
            "and",
            frozenset(
                [("==", ("field_2",), "value_2"), ("==", ("field_1",), "value_1")]
            ),
        ),
    )


def test_db_operations(database):
    item = {"id": "1", "name": "item"}

    assert database.get(**item) is None

    database.get_or_create(**item)
    assert database.get(**item) == item

    filtered = database.filter(id="1")
    assert isinstance(filtered, list)
    assert len(filtered) == 1

    result = filtered[0]
    assert isinstance(result, dict)
    assert result == item

    assert database.all() == [item]
    assert database.filter()

    item["hello"] = "world"
    database.upsert(item)
    assert database.get(**item) == item



import pandas as pd

from entropic.sources import Sample


def test_sample_data_source():
    data_field = {
        "file_path": "tests/mocks/kinematic1.csv",
        "raw": pd.read_csv("tests/mocks/kinematic1.csv"),
    }
    sample = Sample(data=data_field)
    assert sample.data
    assert sample.data.file_path == data_field["file_path"]
    assert not sample.data.raw.empty



from typing import ClassVar, TypeAlias

from pydantic import BaseModel, Field

from entropic.db import default_database

from entropic.sources.sample import Sample


class Iteration(BaseModel):
    database: ClassVar = default_database()
    sample: ClassVar[TypeAlias] = Sample

    samples: list[sample] = Field(default_factory=list)
    source_path: str

    @classmethod
    def get_or_create(cls, **kwargs):
        # TODO: this should be done automatically by the database
        return cls(**cls.database.get_or_create(**kwargs))

    def save(self):
        return self.database.upsert(
            self.model_dump(),
            key={"key": "source_path", "value": self.source_path},
        )

    def upsert_sample(self, sample):
        try:
            if index := self.samples.index(sample):
                self.samples[index] = sample
        except ValueError:
            self.samples.append(sample)
        return sample



from typing import Sequence
from entropic.sources import Iteration
from entropic.db import default_database


class Results:
    database = default_database()
    iteration = Iteration

    def _load(self, document_list: Sequence[dict]) -> Sequence[Iteration]:
        return [self.iteration.model_validate(document) for document in document_list]

    @property
    def all(self) -> Sequence[Iteration]:
        return self._load(self.database.all())
